unit uPlugin;
{$INCLUDE 'Plugin.inc'}

interface

uses
  Windows, SysUtils, Classes,
  {$INCLUDE 'DelphiFarUnits.inc'} ,
  {$IFDEF DEBUG_LOGS} uLog, {$ENDIF}
  uFilter,
  uFilterListUI,
  uOptions,
  uSystem, uPluginConsts, uMessages;

type
  EFilterPluginError = class(EFarPluginError);

  TPlugin = class(TFarPlugin)
  private
  protected
    function PluginID: TGUID; override;
    function GetPluginTitle(out Description, Author: string): string; override;
    procedure GetPluginInfo(var Flags: TPluginFlags; var MainMenu, ConfigMenu, DiskMenu: TFarPlugin.TFarPluginMenu; var CommandPrefix: string); override;
    function GetCommonString(ID: TFarPlugin.TCommonString): string; override;
    procedure DoLog(const Msg: string); override;
    function PluginConfig(GUID: PGUID; MenuItem: TIntPtr): boolean; override;
    function LoadFilterList: TFilterList;
    procedure SaveFilterList(List: TFilterList);
    procedure AddUnconfiguredBuiltInFilters(List: TFilterList);
    function SelectFilter(const FileName, Content: string): TFilter;
    procedure GetEditorInfo(EditorID: TIntPtr; out EditorInfo: TEditorInfo; out FileName, Content: string);
    function ReplaceAllContent(EditorID: TIntPtr; const Content: string): boolean;
    function ReplaceSelectedContent(EditorID: TIntPtr; const Content: string): boolean;
    function InsertNewContent(EditorID: TIntPtr; const Content: string): boolean;
    function Execute(OpenFrom: DWORD; GUID: PGUID; Data: TIntPtr): THandle; override;
    function ProcessEditor: boolean; overload;
    function ProcessEditor(Filter: TFilter): boolean; overload;
    function ProcessEditor(EditorID: TIntPtr): boolean; overload;
    function ProcessEditor(EditorID: TIntPtr; Filter: TFilter): boolean; overload;
    function ProcessMacroCall(const Params: array of Variant): boolean;
  public
    constructor Create;
    destructor Destroy; override;
    function CreateFilterList(WithDefaultFilters: boolean): TFilterList;
    procedure PopulateDefaultFilterList(List: TFilterList);
  end;

implementation

uses
  ufAuto;

const
  PLUGIN_TITLE = 'Filters';
  PLUGIN_DESCRIPTION = 'Filters';
  PLUGIN_AUTHOR = 'Pepak, https://www.pepak.net';

{ TPlugin }

procedure TPlugin.AddUnconfiguredBuiltInFilters(List: TFilterList);
var
  Priority: TFilter.TPriority;
  Filter: TFilter;
  i, j: integer;
  AlreadyPresent: boolean;
begin
  for Priority := High(Priority) downto Low(Priority) do
  begin
    if Priority = prNone then
      Continue;
    for i := 0 to Pred(TRegisteredFilters.Count) do
    begin
      if not TRegisteredFilters.Items[i].AllowUnconfigured then
        Continue;
      if TRegisteredFilters.Items[i].Priority = Priority then
      begin
        AlreadyPresent := False;
        for j := 0 to Pred(List.Count) do
        begin
          if List[j].ClassType = TRegisteredFilters.Items[i] then
          begin
            AlreadyPresent := True;
            Break;
          end;
        end;
        if not AlreadyPresent then
        begin
          Filter := TRegisteredFilters.Items[i].Create;
          List.Add(Filter);
        end;
      end;
    end;
  end;
end;

constructor TPlugin.Create;
begin
  inherited;
end;

function TPlugin.CreateFilterList(WithDefaultFilters: boolean): TFilterList;
var
  List: TFilterList;
begin
  List := LoadFilterList;
  try
    if WithDefaultFilters then
      AddUnconfiguredBuiltInFilters(List);
    Result := List;
    List := nil;
  finally
    FreeAndNil(List);
  end;
end;

destructor TPlugin.Destroy;
begin
  inherited;
end;

procedure TPlugin.DoLog(const Msg: string);
begin
  {$IFDEF DEBUG_LOGS}
  uLog.Log(Msg);
  {$ENDIF}
end;

function TPlugin.Execute(OpenFrom: DWORD; GUID: PGUID; Data: TIntPtr): THandle;
begin
  Result := {$IFDEF FAR3_UP} 0 {$ELSE} INVALID_HANDLE_VALUE {$ENDIF} ;
  if OpenFrom = OPEN_EDITOR then
  begin
    ProcessEditor;
  end;
  {$IFDEF FAR3_UP}
  if OpenFrom = OPEN_FROMMACRO_ then
  begin
    if ProcessMacroCall(TFarMacro.MacroParamsToDelphi(Data)) then
      Result := 1;
  end;
  {$ENDIF}
end;

function TPlugin.GetCommonString(ID: TFarPlugin.TCommonString): string;
begin
  case ID of
    TFarPlugin.TCommonString.MTitle:
      Result := GetMsg(MTitle);
    TFarPlugin.TCommonString.MExceptionOccurredIn:
      Result := GetMsg(MExceptionOccurredIn);
    TFarPlugin.TCommonString.MOk:
      Result := GetMsg(MOk);
    TFarPlugin.TCommonString.MCancel:
      Result := GetMsg(MCancel);
    TFarPlugin.TCommonString.MYes:
      Result := GetMsg(MYes);
    TFarPlugin.TCommonString.MNo:
      Result := GetMsg(MNo);
    else
      Result := inherited GetCommonString(ID);
  end;
end;

procedure TPlugin.GetEditorInfo(EditorID: TIntPtr; out EditorInfo: TEditorInfo; out FileName, Content: string);
begin
  if not TFarEditor.GetInfo(EditorID, EditorInfo) then
    raise EFilterPluginError.Create(GetMsg(MFailedToReadEditorInfo));
  if not TFarEditor.GetFileName(EditorID, FileName) then
    raise EFilterPluginError.Create(GetMsg(MFailedToReadEditorFileName));
  if EditorInfo.BlockType = BTYPE_NONE then
  begin
    if not TFarEditor.GetContent(EditorID, Content) then
      raise EFilterPluginError.Create(GetMsg(MFailedToReadEditorContent));
  end
  else
  begin
    if not TFarEditor.GetSelection(EditorID, Content) then
      raise EFilterPluginError.Create(GetMsg(MFailedToReadEditorContent));
  end;
end;

procedure TPlugin.GetPluginInfo(var Flags: TPluginFlags; var MainMenu, ConfigMenu, DiskMenu: TFarPlugin.TFarPluginMenu; var CommandPrefix: string);
begin
  Flags := Flags or PF_EDITOR or PF_DISABLEPANELS;
  SetLength(MainMenu, 1);
  MainMenu[0].Title := GetMsg(MTitle);
  MainMenu[0].ID := PLUGIN_MENU_MAIN;
  SetLength(ConfigMenu, 1);
  ConfigMenu[0].Title := GetMsg(MTitle);
  ConfigMenu[0].ID := PLUGIN_MENU_CONFIG;
end;

function TPlugin.GetPluginTitle(out Description, Author: string): string;
begin
  Result := PLUGIN_TITLE;
  Description := PLUGIN_DESCRIPTION;
  Author := PLUGIN_AUTHOR;
end;

function TPlugin.InsertNewContent(EditorID: TIntPtr; const Content: string): boolean;
begin
  if not TFarEditor.InsertText(EditorID, Content) then
    raise EFilterPluginError.Create(GetMsg(MFailedToInsertNewContent));
  Result := True;
end;

function TPlugin.LoadFilterList: TFilterList;
var
  FilterKeys: TStringList;
  Config: TOptions;
  FilterConfig: TFilterOptions;
  FailedFilters: TStringList;
  FilterKey: string;
  Filter: TFilter;
  i: Integer;
begin
  Result := TFilterList.Create(True);
  try
    FilterKeys := nil;
    try
      Config := TOptions.Create;
      try
        FilterKeys := Config.GetFilterKeys;
      finally
        FreeAndNil(Config);
      end;
      FilterConfig := TFilterOptions.Create;
      try
        FilterConfig.MustExist := True;
        FailedFilters := TStringList.Create;
        try
          for i := 0 to Pred(FilterKeys.Count) do
          begin
            try
              FilterKey := FilterKeys[i];
              FilterConfig.Key := FilterKey;
              Filter := TFilter.CreateFromConfig(FilterConfig);
              Result.Add(Filter);
            except
              on E: Exception do
              begin
                if FilterKey <> '' then
                  FailedFilters.Add(FilterKey + ': ' + E.ClassName + ': ' + E.Message);
              end;
            end;
          end;
          if FailedFilters.Count > 0 then
          begin
            FailedFilters.Insert(0, GetMsg(MTitle));
            FailedFilters.Insert(1, GetMsg(MFiltersFailedToLoad));
            FailedFilters.Add(GetMsg(MOk));
            ShowMessage(FailedFilters, FMSG_WARNING, 1);
          end;
        finally
          FreeAndNil(FailedFilters);
        end;
      finally
        FreeAndNil(FilterConfig);
      end;
    finally
      FreeAndNil(FilterKeys);
    end;
  except
    FreeAndNil(Result);
    Raise;
  end;
end;

function TPlugin.PluginConfig(GUID: PGUID; MenuItem: TIntPtr): boolean;
var
  List: TFilterList;
  Dlg: TFilterListDialog;
begin
  Result := False;
  List := CreateFilterList(False);
  try
    if List.Count <= 0 then
      PopulateDefaultFilterList(List);
    Dlg := TFilterListDialog.Create;
    try
      Dlg.List := List;
      if Dlg.Execute then
      begin
        SaveFilterList(List);
      end;
    finally
      FreeAndNil(Dlg);
    end;
  finally
    FreeAndNil(List);
  end;
end;

function TPlugin.PluginID: TGUID;
begin
  Result := PLUGIN_GUID;
end;

procedure TPlugin.PopulateDefaultFilterList(List: TFilterList);
var
  Priority: TFilter.TPriority;
  Filter: TFilter;
  i: integer;
begin
  for Priority := High(Priority) downto Low(Priority) do
  begin
    if Priority = prNone then
      Continue;
    for i := 0 to Pred(TRegisteredFilters.Count) do
    begin
      if TRegisteredFilters.Items[i].Priority = Priority then
      begin
        Filter := TRegisteredFilters.Items[i].Create;
        List.Add(Filter);
      end;
    end;
  end;
end;

function TPlugin.ProcessEditor: boolean;
begin
  Result := ProcessEditor(-1);
end;

function TPlugin.ProcessEditor(Filter: TFilter): boolean;
begin
  Result := ProcessEditor(-1, Filter);
end;

function TPlugin.ProcessEditor(EditorID: TIntPtr): boolean;
begin
  Result := ProcessEditor(EditorID, nil);
end;

function TPlugin.ProcessEditor(EditorID: TIntPtr; Filter: TFilter): boolean;
var
  EditorInfo: TEditorInfo;
  FileName: string;
  Content: string;
  CanUndo: boolean;
  ActiveFilter: TFilter;
begin
  Result := False;
  // Get editor's content
  GetEditorInfo(EditorID, EditorInfo, FileName, Content);
  if EditorInfo.BlockType = BTYPE_COLUMN then
    raise EFilterPluginError.Create(GetMsg(MColumnBlocksNotSupported));
  // Find a suitable filter, if none was pre-selected
  if Filter = nil then
  begin
    ActiveFilter := SelectFilter(FileName, Content);
    if ActiveFilter = nil then
      Exit;
  end
  else
  begin
    ActiveFilter := Filter;
  end;
  ActiveFilter := ActiveFilter.Clone;
  try
    // Allow the user to override the settings, if needed
    if ActiveFilter.EditBeforeUse then
    begin
      if not ActiveFilter.ConfigureBeforeUse then
        Abort;
    end;
    CanUndo := TFarEditor.BeginUndoBlock(EditorID);
    try
      // Apply the ActiveFilter to the content
      Content := ActiveFilter.Execute(FileName, Content);
      // Write the content back into the editor
      try
        if ActiveFilter.KeepInputData then
        begin
          Result := InsertNewContent(EditorID, Content);
        end
        else
        if EditorInfo.BlockType = BTYPE_NONE then
        begin
          Result := ReplaceAllContent(EditorID, Content);
        end
        else
        begin
          Result := ReplaceSelectedContent(EditorID, Content);
        end;
      finally
        if CanUndo then
          TFarEditor.EndUndoBlock(EditorID);
      end;
    finally
      if CanUndo and (not Result) then
        TFarEditor.Undo(EditorID);
    end;
  finally
    FreeAndNil(ActiveFilter);
  end;
end;

function TPlugin.ProcessMacroCall(const Params: array of Variant): boolean;
var
  FunctionName: string;
  FilterID: string;
  FilterList: TFilterList;
  FilterClass: TFilterClass;
  FilterName: string;
  Filter: TFilter;
  i: Integer;
begin
  Result := False;
  Filter := nil;
  try
    if Length(Params) > 0 then
    begin
      FunctionName := Params[0];
      if FunctionName = 'execute' then
      begin
        if Length(Params) > 1 then
        begin
          FilterID := Params[1];
          if TRegisteredFilters.Find(FilterID, FilterClass) then
            Filter := FilterClass.Create
          else if FilterID <> '' then
            raise EFilterPluginError.CreateFmt(GetMsg(MUnknownFilter), [FilterID]);
        end;
        if ProcessEditor(Filter) then
          Result := True;
      end
      else if FunctionName = 'execute-by-name' then
      begin
        FilterName := '';
        if Length(Params) > 1 then
        begin
          FilterName := Params[1];
          FilterList := CreateFilterList(True);
          try
            for i := 0 to Pred(FilterList.Count) do
            begin
              if FilterList[i].Description = FilterName then
              begin
                Filter := FilterList[i].Clone;
                Break;
              end;
            end;
          finally
            FreeAndNil(FilterList);
          end;
        end;
        if Filter = nil then
          raise EFilterPluginError.CreateFmt(GetMsg(MUnknownFilter), [FilterName]);
        if ProcessEditor(Filter) then
          Result := True;
      end;
    end;
  finally
    FreeAndNil(Filter);
  end;
end;

function TPlugin.ReplaceAllContent(EditorID: TIntPtr; const Content: string): boolean;
begin
  TFarEditor.SetCursorPosition(EditorID, 0, 0);
  if not TFarEditor.SelectAll(EditorID) then
    raise EFilterPluginError.Create(GetMsg(MFailedToSelectAllContent));
  Result := ReplaceSelectedContent(EditorID, Content);
end;

function TPlugin.ReplaceSelectedContent(EditorID: TIntPtr; const Content: string): boolean;
var
  EditorInfo: TEditorInfo;
  Line, TopLine: integer;
  DummyLine: string;
  SelStart, SelLength: TIntPtr;
begin
  if not TFarEditor.GetInfo(EditorID, EditorInfo) then
    raise EFilterPluginError.Create(GetMsg(MFailedToReadEditorInfo));
  if EditorInfo.BlockType = BTYPE_COLUMN then
    raise EFilterPluginError.Create(GetMsg(MColumnBlocksNotSupported));
  if TFarEditor.GetString(EditorID, EditorInfo.BlockStartLine, DummyLine, SelStart, SelLength) and (SelStart > 0) then
    TFarEditor.SetCursorPosition(EditorID, EditorInfo.BlockStartLine, SelStart - 1);
  if EditorInfo.BlockType <> BTYPE_NONE then
    if not TFarEditor.DeleteSelection(EditorID) then
      raise EFilterPluginError.Create(GetMsg(MFailedToDeleteOldContent));
  Result := InsertNewContent(EditorID, Content);
  if EditorInfo.BlockType = BTYPE_NONE then
  begin
    Line := EditorInfo.CurLine;
    TopLine := EditorInfo.TopScreenLine;
  end
  else
  begin
    Line := EditorInfo.BlockStartLine;
    if Line >= 5 then
      TopLine := Line - 5
    else
      TopLine := 0;
  end;
  TFarEditor.SetCursorPosition(EditorID, Line, EditorInfo.CurPos, EditorInfo.CurTabPos, TopLine, EditorInfo.LeftPos, EditorInfo.Overtype);
end;

procedure TPlugin.SaveFilterList(List: TFilterList);
var
  Config: TOptions;
  FilterConfig: TFilterOptions;
  i: Integer;
begin
  Config := TOptions.Create;
  try
    Config.DeleteAllFilters;
  finally
    FreeAndNil(Config);
  end;
  FilterConfig := TFilterOptions.Create;
  try
    FilterConfig.MustExist := False;
    for i := 0 to Pred(List.Count) do
    begin
      FilterConfig.Key := Format('%6.6d', [i]);
      List[i].SaveToConfig(FilterConfig);
    end;
  finally
    FreeAndNil(FilterConfig);
  end;
end;

function TPlugin.SelectFilter(const FileName, Content: string): TFilter;
var
  FilterList: TFilterList;
  Menu: TStringList;
  SupportedFiltersCount: integer;
  ItemIndex: integer;
  i: integer;
begin
  Result := nil;
  FilterList := CreateFilterList(True);
  try
    // Keep only the supported filters
    SupportedFiltersCount := 0;
    for i := Pred(FilterList.Count) downto 0 do
    begin
      if FilterList[i].IsSupported(FileName, Content) then
      begin
        Inc(SupportedFiltersCount);
      end
      else if FilterList[i] is TFilterAutoSelect then
      begin
        FilterList.Delete(i);
      end
      else
      begin
        FilterList.Move(i, Pred(FilterList.Count));
      end;
    end;
    // If no filter is found, report an error
    if FilterList.Count <= 0 then
      raise EFilterPluginError.Create(GetMsg(MNoApplicableFilterFound));
    // If only one filter remains, use it
    if FilterList.Count = 1 then
    begin
      Result := FilterList[0].Clone;
    end
    // Else let user choose a filter
    else
    begin
      Menu := TStringList.Create;
      try
        for i := 0 to Pred(FilterList.Count) do
        begin
          if (i = SupportedFiltersCount) and (SupportedFiltersCount > 0) then
            Menu.Add(TFarModule.MENU_SEPARATOR);
          if FilterList[i].Description <> '' then
            Menu.Add(FilterList[i].Description)
          else
            Menu.Add(FilterList[i].Name);
        end;
        if Self.ShowMenu(PLUGIN_MENU_FILTERS, GetMsg(MSelectFilter), '', Menu, ItemIndex) then begin
          if (ItemIndex >= SupportedFiltersCount) and (SupportedFiltersCount > 0) then
            Dec(ItemIndex);
          Result := FilterList[ItemIndex].Clone;
        end;
      finally
        FreeAndNil(Menu);
      end;
    end;
  finally
    FreeAndNil(FilterList);
  end;
end;

initialization
  TFarPlugin.InitInstance(TPlugin.Create);

end.
